# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
property property_file_split: PropertyFileSplitModel using property_regex:
    property fy_code
"""

import abc
import re
from functools import cached_property

from fy_library.constants import (
    FY_ENTITY_REGEX_STRING,
    PYTHON_MULTI_ENTITY_REGEX_STRING,
)
from fy_library.mixins.property.property_file_split.abc_fy import PropertyFileSplitModel


from fy_library.mixins.property.fy_code.abc_fy import (
    FyCode_PropertyMixin_ABC,
)

_PROPERTY_REGEX = re.compile(
    rf"(?P<property_annotation>@cached)?\s*"
    rf"property\s+(?P<property_name>{FY_ENTITY_REGEX_STRING})"
    rf"\s*:\s*(?P<return_type>{PYTHON_MULTI_ENTITY_REGEX_STRING})\s*"
    rf"using\s+(?P<implementation_name>{FY_ENTITY_REGEX_STRING})\s*:\s*\n"
)


# fy:start ===>>>
class PropertyFileSplit_UsingPropertyRegex_PropertyMixin(
    # Property_mixins
    FyCode_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _property_file_split(self) -> PropertyFileSplitModel:
        # fy:end <<<===
        property_file_split = _PROPERTY_REGEX.split(self._fy_code)

        assert (
            len(property_file_split) == 6
        ), f"Property file split length {len(property_file_split)} is invalid"

        property_file_split_model = PropertyFileSplitModel(
            user_imports=property_file_split[0],
            property_name=property_file_split[2],
            implementation_name=property_file_split[4],
            property_type=property_file_split[3],
            mixins=property_file_split[5],
        )

        return property_file_split_model



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

import re

from pydantic import BaseModel

_PASCAL_CASE_REGEX = re.compile("(?<=.)(?=[A-Z])")


class PythonEntityName(BaseModel):
    snake_case: str
    pascal_case: str

    @classmethod
    def from_snake_case(cls, snake_case_name: str) -> "PythonEntityName":
        snake_case_phrases = snake_case_name.split("__")
        pascal_case_phrases = [
            "".join(
                snake_case_word.capitalize()
                for snake_case_word in snake_case_phrase.split("_")
            )
            for snake_case_phrase in snake_case_phrases
        ]
        pascal_case_name = "_".join(pascal_case_phrases)
        return cls(
            snake_case=snake_case_name,
            pascal_case=pascal_case_name,
        )

    @classmethod
    def from_pascal_case(cls, pascal_case_name: str) -> "PythonEntityName":
        pascal_case_phrases = pascal_case_name.split("_")
        snake_case_phrases = [
            # TODO: make only the first letter lowercase
            "_".join(
                pascal_case_word.lower()
                for pascal_case_word in _PASCAL_CASE_REGEX.split(pascal_case_phrase)
            )
            for pascal_case_phrase in pascal_case_phrases
        ]
        snake_case_name = "__".join(snake_case_phrases)
        return cls(
            snake_case=snake_case_name,
            pascal_case=pascal_case_name,
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from typing import List


flow AbstractMethodImportsFlow -> List[str]:
    property import_abc using constant
"""

from typing import List

from fy_core.base.flow_base import FlowBase
from fy_library.mixins.property.imports.import_abc__using_constant_fy import (
    ImportAbc_UsingConstant_PropertyMixin,
)


# fy:start ===>>>
class AbstractMethodImportsFlow_Flow(
    # Property Mixins
    ImportAbc_UsingConstant_PropertyMixin,
    # Base
    FlowBase[List[str]],
):
    def __call__(self) -> List[str]:
        # fy:end <<<===
        return self._import_abc



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from typing import List


property mixin_lines: List[str] using property_file_split:
    property property_file_split
"""

import abc
from functools import cached_property
from typing import List

from fy_library.mixins.property.property_file_split.abc_fy import (
    PropertyFileSplit_PropertyMixin_ABC,
)


# fy:start ===>>>
class MixinLines_UsingPropertyFileSplit_PropertyMixin(
    # Property_mixins
    PropertyFileSplit_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _mixin_lines(self) -> List[str]:
        # fy:end <<<===
        return self._property_file_split.mixins.split("\n")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from typing import List


property user_imports_from_mixins: List[str] using property_setter_mixins:
    property parsed_fy_py_files_map_by_key
    property property_setter_mixins
"""

import abc
from functools import cached_property
from typing import List, Set

from fy_library.mixins.property.parsed_fy_py_files_map_by_key.abc_fy import (
    ParsedFyPyFilesMapByKey_PropertyMixin_ABC,
)
from fy_library.mixins.property.property_setter_mixins.abc_fy import (
    PropertySetterMixins_PropertyMixin_ABC,
)


# fy:start ===>>>
class UserImportsFromMixins_UsingPropertySetterMixins_PropertyMixin(
    # Property_mixins
    ParsedFyPyFilesMapByKey_PropertyMixin_ABC,
    PropertySetterMixins_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _user_imports_from_mixins(self) -> List[str]:
        # fy:end <<<===
        user_imports_split = self._parsed_fy_py_files_map_by_key
        user_imports: Set[str] = set()
        for property_setter in self._property_setter_mixins:
            user_imports.update(
                [
                    user_import
                    for user_import in user_imports_split[
                        property_setter.property_name.snake_case
                    ].user_imports.split("\n")
                    if user_import != ""
                ]
            )

        return list(user_imports)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from typing import List
from fy_library.domain.parsed_fy_py_file import ParsedFyPyFile


property required_property_setters_fy_py: List[ParsedFyPyFile]
"""

from fy_library.domain.parsed_fy_py_file import ParsedFyPyFile
from typing import List
import abc


# fy:start ===>>>
class RequiredPropertySettersFyPy_PropertyMixin_ABC(abc.ABC):
    @property
    @abc.abstractmethod
    def _required_property_setters_fy_py(self) -> List[ParsedFyPyFile]:
        raise NotImplementedError()
        # fy:end <<<===



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from typing import List


flow FlowImports -> List[str]:
    property property_mixins using setter
    property method_mixins using setter
    property parsed_fy_py_files_map_by_key using setter
    property mixin_import_map using setter
    property property_setter_mixins using property_mixins
    property user_imports_from_mixins using property_setter_mixins
    property import_any using property_setters_exists
    property import_flow_base using constant
    property property_mixins_import using property_mixins_and_mixin_import_map
    property method_mixins_import using method_mixins_and_mixin_import_map

"""

from typing import List, Any, Dict

from fy_core.base.flow_base import FlowBase
from fy_library.domain.fy_py_template_models import PropertyMixinModel, MethodMixinModel
from fy_library.domain.parsed_fy_py_file import ParsedFyPyFile
from fy_library.mixins.property.imports.import_any__using_property_setters_exists_fy import (
    ImportAny_UsingPropertySettersExists_PropertyMixin,
)
from fy_library.mixins.property.imports.import_flow_base__using_constant_fy import (
    ImportFlowBase_UsingConstant_PropertyMixin,
)
from fy_library.mixins.property.imports.property_mixins_imports__using_property_mixins_and_mixin_import_map_fy import (
    PropertyMixinsImport_UsingPropertyMixinsAndMixinImportMap_PropertyMixin,
)
from fy_library.mixins.property.method_mixins.using_setter import (
    MethodMixins_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.mixin_import_map.using_setter import (
    MixinImportMap_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.parsed_fy_py_files_map_by_key.using_setter import (
    ParsedFyPyFilesMapByKey_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.property_mixins.using_setter import (
    PropertyMixins_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.property_setter_mixins.using_property_mixin_fy import (
    PropertySetterMixins_UsingPropertyMixins_PropertyMixin,
)

from fy_library.mixins.property.imports.method_mixins_imports__using_method_mixins_and_mixin_import_map_fy import (
    MethodMixinsImport_UsingMethodMixinsAndMixinImportMap_PropertyMixin,
)

from fy_library.mixins.property.user_imports_from_property_mixins.using_property_setter_imports_fy import (
    UserImportsFromMixins_UsingPropertySetterMixins_PropertyMixin,
)


# fy:start ===>>>
class FlowImports_Flow(
    # Property Mixins
    PropertyMixins_UsingSetter_PropertyMixin,
    MethodMixins_UsingSetter_PropertyMixin,
    ParsedFyPyFilesMapByKey_UsingSetter_PropertyMixin,
    MixinImportMap_UsingSetter_PropertyMixin,
    PropertySetterMixins_UsingPropertyMixins_PropertyMixin,
    UserImportsFromMixins_UsingPropertySetterMixins_PropertyMixin,
    ImportAny_UsingPropertySettersExists_PropertyMixin,
    ImportFlowBase_UsingConstant_PropertyMixin,
    PropertyMixinsImport_UsingPropertyMixinsAndMixinImportMap_PropertyMixin,
    MethodMixinsImport_UsingMethodMixinsAndMixinImportMap_PropertyMixin,
    # Base
    FlowBase[List[str]],
):
    def __init__(
        self,
        *args: Any,
        property_mixins: List[PropertyMixinModel],
        method_mixins: List[MethodMixinModel],
        parsed_fy_py_files_map_by_key: Dict[str, ParsedFyPyFile],
        mixin_import_map: Dict[str, str],
        **kwargs: Any,
    ):
        self._property_mixins = property_mixins
        self._method_mixins = method_mixins
        self._parsed_fy_py_files_map_by_key = parsed_fy_py_files_map_by_key
        self._mixin_import_map = mixin_import_map
        super().__init__(*args, **kwargs)

    def __call__(self) -> List[str]:
        # fy:end <<<===
        return (
            self._user_imports_from_mixins
            + self._import_any
            + self._import_flow_base
            + self._property_mixins_import
            + self._method_mixins_import
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from typing import Dict


property mixin_import_map: Dict[str, str] using parsed_fy_py_files:
    property parsed_fy_py_files
    property required_property_setters_fy_py
    property project_root_folder
"""

import abc
from functools import cached_property
from typing import Dict

from fy_library.domain.parsed_fy_py_file import ParsedFyPyFile
from fy_library.mixins.property.parsed_fy_py_files.abc_fy import (
    ParsedFyPyFiles_PropertyMixin_ABC,
)
from fy_library.mixins.property.project_root_folder.abc_fy import (
    ProjectRootFolder_PropertyMixin_ABC,
)
from fy_library.mixins.property.required_property_setters_fy_py.abc_fy import (
    RequiredPropertySettersFyPy_PropertyMixin_ABC,
)


# fy:start ===>>>
class MixinImportMap_UsingParsedFyPyFiles_PropertyMixin(
    # Property_mixins
    ParsedFyPyFiles_PropertyMixin_ABC,
    RequiredPropertySettersFyPy_PropertyMixin_ABC,
    ProjectRootFolder_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _mixin_import_map(self) -> Dict[str, str]:
        # fy:end <<<===
        mixin_import_map = {
            parsed_fy_py_file.template_model.entity_key: self.__parsed_file_python_import(
                parsed_fy_py_file
            )
            for parsed_fy_py_file in self._parsed_fy_py_files
            + self._required_property_setters_fy_py
        }
        return mixin_import_map

    def __parsed_file_python_import(self, parsed_fy_py_file: ParsedFyPyFile) -> str:
        relative_file_folder_path = parsed_fy_py_file.file_path.parent.relative_to(
            self._project_root_folder
        )
        file_name = parsed_fy_py_file.file_path.stem
        python_file_path = ".".join(relative_file_folder_path.parts + (file_name,))
        return f"from {python_file_path} import (\n{' ' * 4}{parsed_fy_py_file.template_model.python_class_name.pascal_case},\n)"



# fy:start ===>>>
from pathlib import Path


class FyPyFileToParse_UsingSetter_PropertyMixin:
    @property
    def _fy_py_file_to_parse(self) -> Path:
        return self.__fy_py_file_to_parse

    @_fy_py_file_to_parse.setter
    def _fy_py_file_to_parse(self, fy_py_file_to_parse: Path) -> None:
        self.__fy_py_file_to_parse = fy_py_file_to_parse
        # fy:end <<<===



# fy:start ===>>>
class PreFyCode_UsingSetter_PropertyMixin:
    @property
    def _pre_fy_code(self) -> str:
        return self.__pre_fy_code

    @_pre_fy_code.setter
    def _pre_fy_code(self, pre_fy_code: str) -> None:
        self.__pre_fy_code = pre_fy_code
        # fy:end <<<===



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
property method_file_split: MethodFileSplitModel using method_regex:
    property fy_code
"""

import abc
import re
from functools import cached_property

from fy_library.constants import (
    FY_ENTITY_REGEX_STRING,
    PYTHON_MULTI_ENTITY_REGEX_STRING,
    PYTHON_ARGUMENTS_REGEX_STRING,
)
from fy_library.mixins.property.fy_code.abc_fy import (
    FyCode_PropertyMixin_ABC,
)
from fy_library.mixins.property.method_file_split.abc_fy import MethodFileSplitModel

_METHOD_STRING_SPLIT_REGEX = re.compile(
    rf"method\s+(?P<method_name>{FY_ENTITY_REGEX_STRING})\s*"
    rf"(?P<arguments>\(({PYTHON_ARGUMENTS_REGEX_STRING})\))?\s+->"
    rf"\s+(?P<return_type>{PYTHON_MULTI_ENTITY_REGEX_STRING})\s+"
    rf"using\s+(?P<implementation_name>{FY_ENTITY_REGEX_STRING})\s*:\s*\n"
)


# fy:start ===>>>
class MethodFileSplit_UsingMethodRegex_PropertyMixin(
    # Property_mixins
    FyCode_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _method_file_split(self) -> MethodFileSplitModel:
        # fy:end <<<===
        method_file_split = _METHOD_STRING_SPLIT_REGEX.split(self._fy_code)

        assert (
            len(method_file_split)
        ) == 7, f"Method file split length {len(method_file_split)} is invalid."

        method_file_split_model = MethodFileSplitModel(
            user_imports=method_file_split[0],
            method_name=method_file_split[1],
            implementation_name=method_file_split[5],
            arguments=method_file_split[3],
            return_type=method_file_split[4],
            mixins=method_file_split[6],
        )

        return method_file_split_model



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from fy_library.domain.parsed_fy_py_file import ParsedFyPyFile


method parse_fy_py_file -> ParsedFyPyFile using fy_file_kind__and__fy_code:
    property fy_py_file_to_parse
    property pre_fy_code
    property fy_code
    property pre_marker_file_content
    property post_marker_file_content
    property fy_file_kind
"""

import abc

from fy_core.base.flow_base import FlowBase
from fy_library.domain.parsed_fy_py_file import ParsedFyPyFile, ParsedFyPyFileKind
from fy_library.flows.parse_abstract_method_fy_code_fy import (
    ParseAbstractMethodFyCode_Flow,
)
from fy_library.flows.parse_abstract_property_fy_code_fy import (
    ParseAbstractPropertyFyCode_Flow,
)
from fy_library.flows.parse_flow_fy_code_fy import ParseFlowFyCode_Flow
from fy_library.flows.parse_method_fy_code_fy import ParseMethodFyCode_Flow
from fy_library.flows.parse_property_fy_code_fy import ParsePropertyFyCode_Flow
from fy_library.mixins.property.fy_code.abc_fy import (
    FyCode_PropertyMixin_ABC,
)
from fy_library.mixins.property.fy_file_kind.abc_fy import (
    FyFileKind_PropertyMixin_ABC,
)
from fy_library.mixins.property.fy_py_file_to_parse.abc_fy import (
    FyPyFileToParse_PropertyMixin_ABC,
)
from fy_library.mixins.property.post_marker_file_content.abc_fy import (
    PostMarkerFileContent_PropertyMixin_ABC,
)
from fy_library.mixins.property.pre_fy_code.abc_fy import PreFyCode_PropertyMixin_ABC
from fy_library.mixins.property.pre_marker_file_content.abc_fy import (
    PreMarkerFileContent_PropertyMixin_ABC,
)


# fy:start ===>>>
class ParseFyPyFile_UsingFyFileKind_And_FyCode_MethodMixin(
    # Property_mixins
    FyPyFileToParse_PropertyMixin_ABC,
    PreFyCode_PropertyMixin_ABC,
    FyCode_PropertyMixin_ABC,
    PreMarkerFileContent_PropertyMixin_ABC,
    PostMarkerFileContent_PropertyMixin_ABC,
    FyFileKind_PropertyMixin_ABC,
    abc.ABC,
):
    def _parse_fy_py_file(self) -> ParsedFyPyFile:
        # fy:end <<<===
        parse_fy_code: FlowBase[ParsedFyPyFile]
        match self._fy_file_kind:
            case ParsedFyPyFileKind.FLOW:
                parse_fy_code = ParseFlowFyCode_Flow(
                    pre_fy_code=self._pre_fy_code,
                    fy_code=self._fy_code,
                    pre_marker_file_content=self._pre_marker_file_content,
                    post_marker_file_content=self._post_marker_file_content,
                    fy_py_file_to_parse=self._fy_py_file_to_parse,
                )
            case ParsedFyPyFileKind.METHOD:
                parse_fy_code = ParseMethodFyCode_Flow(
                    pre_fy_code=self._pre_fy_code,
                    fy_code=self._fy_code,
                    pre_marker_file_content=self._pre_marker_file_content,
                    post_marker_file_content=self._post_marker_file_content,
                    fy_py_file_to_parse=self._fy_py_file_to_parse,
                )
            case ParsedFyPyFileKind.ABSTRACT_METHOD:
                parse_fy_code = ParseAbstractMethodFyCode_Flow(
                    pre_fy_code=self._pre_fy_code,
                    fy_code=self._fy_code,
                    pre_marker_file_content=self._pre_marker_file_content,
                    post_marker_file_content=self._post_marker_file_content,
                    fy_py_file_to_parse=self._fy_py_file_to_parse,
                )
            case ParsedFyPyFileKind.ABSTRACT_PROPERTY:
                parse_fy_code = ParseAbstractPropertyFyCode_Flow(
                    pre_fy_code=self._pre_fy_code,
                    fy_code=self._fy_code,
                    pre_marker_file_content=self._pre_marker_file_content,
                    post_marker_file_content=self._post_marker_file_content,
                    fy_py_file_to_parse=self._fy_py_file_to_parse,
                )
            case ParsedFyPyFileKind.PROPERTY:
                parse_fy_code = ParsePropertyFyCode_Flow(
                    pre_fy_code=self._pre_fy_code,
                    fy_code=self._fy_code,
                    pre_marker_file_content=self._pre_marker_file_content,
                    post_marker_file_content=self._post_marker_file_content,
                    fy_py_file_to_parse=self._fy_py_file_to_parse,
                )
            case _:
                raise NotImplementedError(
                    f"Unimplemented fy file kind parser for {self._fy_file_kind}"
                )

        return parse_fy_code()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
from pathlib import Path
from typing import Literal

from pydantic import BaseModel
from enum import Enum

from fy_library.domain.fy_py_template_models import (
    FlowTemplateModel,
    BaseTemplateModel,
    MethodTemplateModel,
    AbstractMethodTemplateModel,
    AbstractPropertyTemplateModel,
    PropertyTemplateModel,
    PropertySetterTemplateModel,
)


class ParsedFyPyFileKind(Enum):
    FLOW = "flow"
    METHOD = "method"
    ABSTRACT_METHOD = "abstract_method"
    ABSTRACT_PROPERTY = "abstract_property"
    PROPERTY = "property"
    PROPERTY_SETTER = "property_setter"


class FyPyFileParts(BaseModel):
    pre_fy_code: str
    fy_code: str
    pre_marker_file_content: str
    post_marker_file_content: str


class ParsedFyPyFile(FyPyFileParts):
    file_type: ParsedFyPyFileKind
    file_path: Path
    user_imports: str
    template_model: BaseTemplateModel


class ParsedFlowFyPyFile(ParsedFyPyFile):
    file_type: Literal[ParsedFyPyFileKind.FLOW] = ParsedFyPyFileKind.FLOW
    template_model: FlowTemplateModel


class ParsedMethodFyPyFile(ParsedFyPyFile):
    file_type: Literal[ParsedFyPyFileKind.METHOD] = ParsedFyPyFileKind.METHOD
    template_model: MethodTemplateModel


class ParsedAbstractMethodFyPyFile(ParsedFyPyFile):
    file_type: Literal[ParsedFyPyFileKind.ABSTRACT_METHOD] = (
        ParsedFyPyFileKind.ABSTRACT_METHOD
    )
    template_model: AbstractMethodTemplateModel


class ParsedAbstractPropertyFyPyFile(ParsedFyPyFile):
    file_type: Literal[ParsedFyPyFileKind.ABSTRACT_PROPERTY] = (
        ParsedFyPyFileKind.ABSTRACT_PROPERTY
    )
    template_model: AbstractPropertyTemplateModel


class ParsedPropertyFyPyFile(ParsedFyPyFile):
    file_type: Literal[ParsedFyPyFileKind.PROPERTY] = ParsedFyPyFileKind.PROPERTY
    template_model: PropertyTemplateModel


class PropertySetterFyPyFile(ParsedFyPyFile):
    file_type: Literal[ParsedFyPyFileKind.PROPERTY_SETTER] = (
        ParsedFyPyFileKind.PROPERTY_SETTER
    )
    template_model: PropertySetterTemplateModel



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from pathlib import Path


property project_root_folder: Path
"""

import abc
from pathlib import Path


# fy:start ===>>>
class ProjectRootFolder_PropertyMixin_ABC(abc.ABC):
    @property
    @abc.abstractmethod
    def _project_root_folder(self) -> Path:
        raise NotImplementedError()
        # fy:end <<<===



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from fy_library.mixins.property.included_mixins.abc_fy import IncludedMixinsModel


property included_mixins: IncludedMixinsModel using mixin_lines:
    property fy_py_file_to_parse
    property mixin_lines
"""

import abc
import re
from functools import cached_property
from typing import List

from fy_library.constants import FY_ENTITY_REGEX_STRING
from fy_library.domain.fy_py_template_models import (
    PropertyMixinModel,
    AbstractPropertyModel,
    AbstractMethodModel,
    MethodMixinModel,
)
from fy_library.domain.python_entity_name import PythonEntityName
from fy_library.mixins.property.fy_py_file_to_parse.abc_fy import (
    FyPyFileToParse_PropertyMixin_ABC,
)
from fy_library.mixins.property.included_mixins.abc_fy import IncludedMixinsModel
from fy_library.mixins.property.mixin_lines.abc_fy import (
    MixinLines_PropertyMixin_ABC,
)

_FLOW_METHOD_REGEX = re.compile(
    rf"^\s+method\s+(?P<method_name>{FY_ENTITY_REGEX_STRING})\s+"
    rf"using\s+(?P<implementation_name>{FY_ENTITY_REGEX_STRING})\s*$"
)

_FLOW_PROPERTY_REGEX = re.compile(
    rf"^\s+property\s+(?P<property_name>{FY_ENTITY_REGEX_STRING})\s+"
    rf"using\s+(?P<implementation_name>{FY_ENTITY_REGEX_STRING})\s*$"
)

_ABSTRACT_PROPERTY_MIXIN_REGEX = re.compile(
    rf"^\s+property\s+(?P<abstract_property_name>{FY_ENTITY_REGEX_STRING})$"
)

_ABSTRACT_METHOD_MIXIN_REGEX = re.compile(
    rf"^\s+method\s+(?P<abstract_method_name>{FY_ENTITY_REGEX_STRING})$"
)


# fy:start ===>>>
class IncludedMixins_UsingMixinLines_PropertyMixin(
    # Property_mixins
    FyPyFileToParse_PropertyMixin_ABC,
    MixinLines_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _included_mixins(self) -> IncludedMixinsModel:
        # fy:end <<<===
        properties: List[PropertyMixinModel] = []
        abstract_properties: List[AbstractPropertyModel] = []
        abstract_methods: List[AbstractMethodModel] = []
        methods: List[MethodMixinModel] = []
        for mixin_line in self._mixin_lines:
            if mixin_line.strip() == "":
                continue

            flow_property_fy_search = _FLOW_PROPERTY_REGEX.search(mixin_line)

            if flow_property_fy_search is not None:
                properties.append(
                    PropertyMixinModel(
                        property_name=PythonEntityName.from_snake_case(
                            flow_property_fy_search.group("property_name")
                        ),
                        implementation_name=PythonEntityName.from_snake_case(
                            flow_property_fy_search.group("implementation_name")
                        ),
                    )
                )
                continue

            declared_abstract_property_mixin = _ABSTRACT_PROPERTY_MIXIN_REGEX.search(
                mixin_line
            )

            if declared_abstract_property_mixin is not None:
                abstract_properties.append(
                    AbstractPropertyModel(
                        property_name=PythonEntityName.from_snake_case(
                            declared_abstract_property_mixin.group(
                                "abstract_property_name"
                            )
                        )
                    )
                )
                continue

            declared_abstract_method_mixin = _ABSTRACT_METHOD_MIXIN_REGEX.search(
                mixin_line
            )

            if declared_abstract_method_mixin is not None:
                abstract_methods.append(
                    AbstractMethodModel(
                        method_name=PythonEntityName.from_snake_case(
                            declared_abstract_method_mixin.group("abstract_method_name")
                        )
                    )
                )
                continue

            flow_method_fy_search = _FLOW_METHOD_REGEX.search(mixin_line)

            if flow_method_fy_search is not None:
                methods.append(
                    MethodMixinModel(
                        method_name=PythonEntityName.from_snake_case(
                            flow_method_fy_search.group("method_name")
                        ),
                        implementation_name=PythonEntityName.from_snake_case(
                            flow_method_fy_search.group("implementation_name")
                        ),
                    )
                )
                continue

            raise ValueError(
                f"Line {mixin_line} in {self._fy_py_file_to_parse} is invalid."
            )

        included_mixins = IncludedMixinsModel(
            abstract_method_mixins=abstract_methods,
            abstract_property_mixins=abstract_properties,
            method_mixins=methods,
            property_mixins=properties,
        )

        return included_mixins



# fy:start ===>>>
from typing import List
from fy_library.domain.fy_py_template_models import MethodMixinModel


class MethodMixins_UsingSetter_PropertyMixin:
    @property
    def _method_mixins(self) -> List[MethodMixinModel]:
        return self.__method_mixins

    @_method_mixins.setter
    def _method_mixins(self, method_mixins: List[MethodMixinModel]) -> None:
        self.__method_mixins = method_mixins
        # fy:end <<<===



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from typing import List


flow PropertyImports -> List[str]:
    property abstract_property_mixins using setter
    property mixin_import_map using setter
    property cached_import using constant
    property import_abc using when_abstract_property_mixins_exists
    property import_abstract_property_mixins using abstract_property_mixin_and_mixin_import_map
"""

from typing import List, Any, Dict

from fy_core.base.flow_base import FlowBase
from fy_library.domain.fy_py_template_models import AbstractPropertyModel
from fy_library.mixins.property.abstract_property_mixins.using_setter import (
    AbstractPropertyMixins_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.imports.cached_import__using_constant_fy import (
    CachedImport_UsingConstant_PropertyMixin,
)
from fy_library.mixins.property.imports.import__abstract_property_mixins__using_abstract_property_mixin_and_mixin_import_map__fy import (
    ImportAbstractPropertyMixins_UsingAbstractPropertyMixinAndMixinImportMap_PropertyMixin,
)
from fy_library.mixins.property.imports.import_abc__using_when_abstract_property_mixins_exists__fy import (
    ImportAbc_UsingWhenAbstractPropertyMixinsExists_PropertyMixin,
)
from fy_library.mixins.property.mixin_import_map.using_setter import (
    MixinImportMap_UsingSetter_PropertyMixin,
)


# fy:start ===>>>
class PropertyImports_Flow(
    # Property Mixins
    AbstractPropertyMixins_UsingSetter_PropertyMixin,
    MixinImportMap_UsingSetter_PropertyMixin,
    CachedImport_UsingConstant_PropertyMixin,
    ImportAbc_UsingWhenAbstractPropertyMixinsExists_PropertyMixin,
    ImportAbstractPropertyMixins_UsingAbstractPropertyMixinAndMixinImportMap_PropertyMixin,
    # Base
    FlowBase[List[str]],
):
    def __init__(
        self,
        *args: Any,
        abstract_property_mixins: List[AbstractPropertyModel],
        mixin_import_map: Dict[str, str],
        **kwargs: Any,
    ):
        self._abstract_property_mixins = abstract_property_mixins
        self._mixin_import_map = mixin_import_map
        super().__init__(*args, **kwargs)

    def __call__(self) -> List[str]:
        # fy:end <<<===
        return (
            self._cached_import
            + self._import_abc
            + self._import_abstract_property_mixins
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
flow GenerateAndSaveFyPyFile_UsingParsedFyPyFile -> None:
    property parsed_fy_py_file using setter
    property mixin_import_map using setter
    property parsed_fy_py_files_map_by_key using setter
    property mixin_imports using parsed_fy_py_file
    property jinja2_template_file_name using parsed_fy_py_file
    property template_model using parsed_fy_py_file
    property generated_fy_py_code using jinja2_templates
    property filtered_mixin_imports using remove_existing_imports
    property mixin_imports_code using filtered_mixin_imports
    property fy_py_file_content using parsed_fy_py_file
    method generate_and_save_fy_py_code using parsed_fy_py_file__and__fy_py_file_content
"""

from typing import Any, Dict

from fy_core.base.flow_base import FlowBase
from fy_library.domain.parsed_fy_py_file import (
    ParsedFyPyFile,
)
from fy_library.mixins.method.generate_and_save_fy_py_code.using_parsed_fy_py_file__and__fy_py_file_content_fy import (
    GenerateAndSaveFyPyCode_UsingParsedFyPyFile_And_FyPyFileContent_MethodMixin,
)
from fy_library.mixins.property.filtered_mixin_imports.using_remove_existing_imports_fy import (
    FilteredMixinImports_UsingRemoveExistingImports_PropertyMixin,
)
from fy_library.mixins.property.fy_py_file_content.using_parsed_fy_py_file_fy import (
    FyPyFileContent_UsingParsedFyPyFile_PropertyMixin,
)
from fy_library.mixins.property.generated_fy_py_code.using_jinja2_templates_fy import (
    GeneratedFyPyCode_UsingJinja2Templates_PropertyMixin,
)
from fy_library.mixins.property.jinja2_template_file_name.using_parsed_fy_py_file_fy import (
    Jinja2TemplateFileName_UsingParsedFyPyFile_PropertyMixin,
)
from fy_library.mixins.property.mixin_import_map.using_setter import (
    MixinImportMap_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.mixin_imports.using_parsed_fy_py_file_fy import (
    MixinImports_UsingParsedFyPyFile_PropertyMixin,
)
from fy_library.mixins.property.mixin_imports_code.using_filtered_mixin_imports_fy import (
    MixinImportsCode_UsingFilteredMixinImports_PropertyMixin,
)
from fy_library.mixins.property.parsed_fy_py_file.using_setter import (
    ParsedFyPyFile_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.parsed_fy_py_files_map_by_key.using_setter import (
    ParsedFyPyFilesMapByKey_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.template_model.using_parsed_fy_py_file_fy import (
    TemplateModel_UsingParsedFyPyFile_PropertyMixin,
)


# fy:start ===>>>
class GenerateAndSaveFyPyFile_UsingParsedFyPyFile_Flow(
    # Property Mixins
    ParsedFyPyFile_UsingSetter_PropertyMixin,
    MixinImportMap_UsingSetter_PropertyMixin,
    ParsedFyPyFilesMapByKey_UsingSetter_PropertyMixin,
    MixinImports_UsingParsedFyPyFile_PropertyMixin,
    Jinja2TemplateFileName_UsingParsedFyPyFile_PropertyMixin,
    TemplateModel_UsingParsedFyPyFile_PropertyMixin,
    GeneratedFyPyCode_UsingJinja2Templates_PropertyMixin,
    FilteredMixinImports_UsingRemoveExistingImports_PropertyMixin,
    MixinImportsCode_UsingFilteredMixinImports_PropertyMixin,
    FyPyFileContent_UsingParsedFyPyFile_PropertyMixin,
    # Method Mixins
    GenerateAndSaveFyPyCode_UsingParsedFyPyFile_And_FyPyFileContent_MethodMixin,
    # Base
    FlowBase[None],
):
    def __init__(
        self,
        *args: Any,
        parsed_fy_py_file: ParsedFyPyFile,
        mixin_import_map: Dict[str, str],
        parsed_fy_py_files_map_by_key: Dict[str, ParsedFyPyFile],
        **kwargs: Any,
    ):
        self._parsed_fy_py_file = parsed_fy_py_file
        self._mixin_import_map = mixin_import_map
        self._parsed_fy_py_files_map_by_key = parsed_fy_py_files_map_by_key
        super().__init__(*args, **kwargs)

    def __call__(self) -> None:
        # fy:end <<<===
        self._generate_and_save_fy_py_code()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
import abc
from typing import List

from pydantic import BaseModel, computed_field

from fy_library.domain.python_entity_name import PythonEntityName


def entity_key(
    mixin_name__snake_case: str, mixin_implementation_name__snake_case: str
) -> str:
    return f"{mixin_name__snake_case}.{mixin_implementation_name__snake_case}"


class PropertyMixinModel(BaseModel):
    property_name: PythonEntityName
    implementation_name: PythonEntityName


class MethodMixinModel(BaseModel):
    method_name: PythonEntityName
    implementation_name: PythonEntityName


class AbstractPropertyModel(BaseModel):
    property_name: PythonEntityName


class AbstractMethodModel(BaseModel):
    method_name: PythonEntityName


class BaseTemplateModel(BaseModel, abc.ABC):
    python_class_name: PythonEntityName

    @property
    @abc.abstractmethod
    def entity_key(self) -> str:
        raise NotImplementedError()


class FlowTemplateModel(BaseTemplateModel):
    flow_name: PythonEntityName
    return_type: str
    properties: List[PropertyMixinModel]
    methods: List[MethodMixinModel]

    @computed_field
    @property
    def entity_key(self) -> str:
        return self.flow_name.snake_case


class MethodTemplateModel(BaseTemplateModel):
    method_name: PythonEntityName
    implementation_name: PythonEntityName
    abstract_property_mixins: List[AbstractPropertyModel]
    abstract_method_mixins: List[AbstractMethodModel]
    arguments: str | None
    return_type: str

    @computed_field
    @property
    def entity_key(self) -> str:
        return entity_key(
            mixin_name__snake_case=self.method_name.snake_case,
            mixin_implementation_name__snake_case=self.implementation_name.snake_case,
        )


class AbstractMethodTemplateModel(BaseTemplateModel):
    abstract_method_name: PythonEntityName
    arguments: str | None
    return_type: str

    @computed_field
    @property
    def entity_key(self) -> str:
        return self.abstract_method_name.snake_case


class AbstractPropertyTemplateModel(BaseTemplateModel):
    abstract_property_name: PythonEntityName
    property_type: str

    @computed_field
    @property
    def entity_key(self) -> str:
        return self.abstract_property_name.snake_case


class FlowTemplateModelWithPropertySetters(FlowTemplateModel):
    property_setters: List[AbstractPropertyTemplateModel]


class PropertyTemplateModel(BaseTemplateModel):
    property_name: PythonEntityName
    implementation_name: PythonEntityName
    abstract_property_mixins: List[AbstractPropertyModel]
    property_type: str

    @computed_field
    @property
    def entity_key(self) -> str:
        return entity_key(
            mixin_name__snake_case=self.property_name.snake_case,
            mixin_implementation_name__snake_case=self.implementation_name.snake_case,
        )


class PropertySetterTemplateModel(BaseTemplateModel):
    property_name: PythonEntityName
    property_type: str

    @computed_field
    @property
    def entity_key(self) -> str:
        return entity_key(
            mixin_name__snake_case=self.property_name.snake_case,
            mixin_implementation_name__snake_case="setter",
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from pathlib import Path
from typing import List


property fy_py_files_to_parse: List[Path]
"""

from pathlib import Path
from typing import List
import abc


# fy:start ===>>>
class FyPyFilesToParse_PropertyMixin_ABC(abc.ABC):
    @property
    @abc.abstractmethod
    def _fy_py_files_to_parse(self) -> List[Path]:
        raise NotImplementedError()
        # fy:end <<<===



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
flow GenerateAndSaveFyPyFiles_UsingRequiredPropertySetters -> None:
    property required_property_setters_fy_py using setter
"""

from typing import Any, List

from fy_core.base.flow_base import FlowBase
from fy_library.domain.parsed_fy_py_file import ParsedFyPyFile
from fy_library.flows.generate_and_save_fy_py_file__using_required_property_setters_fy import (
    GenerateAndSaveFyPyFile_UsingRequiredPropertySetters_Flow,
)
from fy_library.mixins.property.required_property_setters_fy_py.using_setter import (
    RequiredPropertySettersFyPy_UsingSetter_PropertyMixin,
)


# fy:start ===>>>
class GenerateAndSaveFyPyFiles_UsingRequiredPropertySetters_Flow(
    # Property Mixins
    RequiredPropertySettersFyPy_UsingSetter_PropertyMixin,
    # Base
    FlowBase[None],
):
    def __init__(
        self,
        *args: Any,
        required_property_setters_fy_py: List[ParsedFyPyFile],
        **kwargs: Any,
    ):
        self._required_property_setters_fy_py = required_property_setters_fy_py
        super().__init__(*args, **kwargs)

    def __call__(self) -> None:
        # fy:end <<<===
        for parsed_fy_py_file in self._required_property_setters_fy_py:
            GenerateAndSaveFyPyFile_UsingRequiredPropertySetters_Flow(
                parsed_fy_py_file=parsed_fy_py_file,
            )()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
property flow_file_split: FlowFileSplitModel using flow_regex:
    property fy_code
"""

import abc
import re
from functools import cached_property

from fy_library.constants import (
    FY_ENTITY_REGEX_STRING,
    PYTHON_MULTI_ENTITY_REGEX_STRING,
)
from fy_library.mixins.property.flow_file_split.abc_fy import FlowFileSplitModel
from fy_library.mixins.property.fy_code.abc_fy import (
    FyCode_PropertyMixin_ABC,
)

_FLOW_STRING_SPLIT_REGEX = re.compile(
    rf"flow\s+(?P<flow_name>{FY_ENTITY_REGEX_STRING})\s+->"
    rf"\s+(?P<return_type>{PYTHON_MULTI_ENTITY_REGEX_STRING}):\s*\n"
)


# fy:start ===>>>
class FlowFileSplit_UsingFlowRegex_PropertyMixin(
    # Property_mixins
    FyCode_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _flow_file_split(self) -> FlowFileSplitModel:
        # fy:end <<<===
        flow_file_split = _FLOW_STRING_SPLIT_REGEX.split(self._fy_code)

        assert (
            len(flow_file_split)
        ) == 4, f"Flow file split length {len(flow_file_split)} is invalid."

        flow_file_split_model = FlowFileSplitModel(
            user_imports=flow_file_split[0],
            flow_name=flow_file_split[1],
            return_type=flow_file_split[2],
            mixins=flow_file_split[3],
        )

        return flow_file_split_model



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from fy_library.domain.parsed_fy_py_file import ParsedFyPyFileKind


property fy_file_kind: ParsedFyPyFileKind
"""

from fy_library.domain.parsed_fy_py_file import ParsedFyPyFileKind
import abc


# fy:start ===>>>
class FyFileKind_PropertyMixin_ABC(abc.ABC):
    @property
    @abc.abstractmethod
    def _fy_file_kind(self) -> ParsedFyPyFileKind:
        raise NotImplementedError()
        # fy:end <<<===



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from typing import List
from fy_library.domain.parsed_fy_py_file import ParsedFyPyFile


property required_property_setters_fy_py: List[ParsedFyPyFile] using parsed_fy_py_files:
    property parsed_fy_py_files
    property parsed_fy_py_files_map_by_key
"""

import abc
from functools import cached_property
from typing import List, cast

from fy_library.domain.fy_py_template_models import (
    PropertySetterTemplateModel,
    AbstractPropertyTemplateModel,
)
from fy_library.domain.parsed_fy_py_file import (
    ParsedFyPyFile,
    PropertySetterFyPyFile,
    ParsedFyPyFileKind,
    ParsedFlowFyPyFile,
)
from fy_library.domain.python_entity_name import PythonEntityName
from fy_library.mixins.property.parsed_fy_py_files.abc_fy import (
    ParsedFyPyFiles_PropertyMixin_ABC,
)
from fy_library.mixins.property.parsed_fy_py_files_map_by_key.abc_fy import (
    ParsedFyPyFilesMapByKey_PropertyMixin_ABC,
)


# fy:start ===>>>
class RequiredPropertySettersFyPy_UsingParsedFyPyFiles_PropertyMixin(
    # Property_mixins
    ParsedFyPyFiles_PropertyMixin_ABC,
    ParsedFyPyFilesMapByKey_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _required_property_setters_fy_py(self) -> List[ParsedFyPyFile]:
        # fy:end <<<===
        required_setters = {
            flow_property.property_name.snake_case: PropertySetterFyPyFile(
                pre_fy_code="",
                fy_code="",
                pre_marker_file_content="",
                post_marker_file_content="",
                file_path=self._parsed_fy_py_files_map_by_key[
                    flow_property.property_name.snake_case
                ].file_path.with_name("using_setter.py"),
                user_imports=self._parsed_fy_py_files_map_by_key[
                    flow_property.property_name.snake_case
                ].user_imports,
                template_model=PropertySetterTemplateModel(
                    property_name=flow_property.property_name,
                    python_class_name=PythonEntityName.from_pascal_case(
                        f"{flow_property.property_name.pascal_case}_UsingSetter_PropertyMixin"
                    ),
                    property_type=cast(
                        AbstractPropertyTemplateModel,
                        self._parsed_fy_py_files_map_by_key[
                            flow_property.property_name.snake_case
                        ].template_model,
                    ).property_type,
                ),
            )
            for parsed_fy_py_file in self._parsed_fy_py_files
            if parsed_fy_py_file.file_type == ParsedFyPyFileKind.FLOW
            for flow_property in cast(
                ParsedFlowFyPyFile, parsed_fy_py_file
            ).template_model.properties
            if (
                flow_property.implementation_name.snake_case == "setter"
                and f"{flow_property.property_name.snake_case}.setter"
                not in self._parsed_fy_py_files_map_by_key
            )
        }

        return list(required_setters.values())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
property pre_fy_code: str using fy_py_file_parts:
    property fy_py_file_parts
"""

import abc
from functools import cached_property

from fy_library.mixins.property.fy_py_file_parts.abc_fy import (
    FyPyFileParts_PropertyMixin_ABC,
)


# fy:start ===>>>
class PreFyCode_UsingFyPyFileParts_PropertyMixin(
    # Property_mixins
    FyPyFileParts_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _pre_fy_code(self) -> str:
        # fy:end <<<===
        return self._fy_py_file_parts.pre_fy_code



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from fy_library.domain.parsed_fy_py_file import ParsedFyPyFile


flow ParseFyPyFile -> ParsedFyPyFile:
    property fy_py_file_to_parse using setter
    property fy_py_file_parts using fy_file_to_parse_docstring
    property pre_fy_code using fy_py_file_parts
    property fy_code using fy_py_file_parts
    property post_marker_file_content using fy_py_file_parts
    property pre_marker_file_content using fy_py_file_parts
    property fy_file_kind using fy_code

    method parse_fy_py_file using fy_file_kind__and__fy_code
"""

from pathlib import Path
from typing import Any

from fy_core.base.flow_base import FlowBase
from fy_library.domain.parsed_fy_py_file import ParsedFyPyFile
from fy_library.mixins.method.parse_fy_py_file.using_fy_file_kind__and__fy_code_fy import (
    ParseFyPyFile_UsingFyFileKind_And_FyCode_MethodMixin,
)
from fy_library.mixins.property.fy_code.using_fy_py_file_parts_fy import (
    FyCode_UsingFyPyFileParts_PropertyMixin,
)
from fy_library.mixins.property.fy_file_kind.using_fy_code_fy import (
    FyFileKind_UsingFyCode_PropertyMixin,
)
from fy_library.mixins.property.fy_py_file_parts.using_fy_file_to_parse_docstring_fy import (
    FyPyFileParts_UsingFyFileToParseDocstring_PropertyMixin,
)
from fy_library.mixins.property.fy_py_file_to_parse.using_setter import (
    FyPyFileToParse_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.post_marker_file_content.using_fy_py_file_parts_fy import (
    PostMarkerFileContent_UsingFyPyFileParts_PropertyMixin,
)
from fy_library.mixins.property.pre_fy_code.using_fy_py_file_parts_fy import (
    PreFyCode_UsingFyPyFileParts_PropertyMixin,
)
from fy_library.mixins.property.pre_marker_file_content.using_fy_py_file_parts_fy import (
    PreMarkerFileContent_UsingFyPyFileParts_PropertyMixin,
)


# fy:start ===>>>
class ParseFyPyFile_Flow(
    # Property Mixins
    FyPyFileToParse_UsingSetter_PropertyMixin,
    FyPyFileParts_UsingFyFileToParseDocstring_PropertyMixin,
    PreFyCode_UsingFyPyFileParts_PropertyMixin,
    FyCode_UsingFyPyFileParts_PropertyMixin,
    PostMarkerFileContent_UsingFyPyFileParts_PropertyMixin,
    PreMarkerFileContent_UsingFyPyFileParts_PropertyMixin,
    FyFileKind_UsingFyCode_PropertyMixin,
    # Method Mixins
    ParseFyPyFile_UsingFyFileKind_And_FyCode_MethodMixin,
    # Base
    FlowBase[ParsedFyPyFile],
):
    def __init__(
        self,
        *args: Any,
        fy_py_file_to_parse: Path,
        **kwargs: Any,
    ):
        self._fy_py_file_to_parse = fy_py_file_to_parse
        super().__init__(*args, **kwargs)

    def __call__(self) -> ParsedFyPyFile:
        # fy:end <<<===
        return self._parse_fy_py_file()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from pathlib import Path
from typing import Any


flow FyPy_Main -> None:
    property folder_to_parse using setter
    property project_root_folder using setter
    property fy_py_files_to_parse using files_discovery
    property parsed_fy_py_files using fy_py_parser
    property parsed_fy_py_files_map_by_key using parsed_fy_py_files
    property required_property_setters_fy_py using parsed_fy_py_files
    property mixin_import_map using parsed_fy_py_files
"""

from pathlib import Path
from typing import Any

from fy_core.base.flow_base import FlowBase
from fy_library.flows.generate_and_save_fy_py_files__using_parsed_fy_py_files_fy import (
    GenerateAndSaveFyPyFiles_UsingParsedFyPyFiles_Flow,
)
from fy_library.flows.generate_and_save_fy_py_files__using_required_property_setters_fy import (
    GenerateAndSaveFyPyFiles_UsingRequiredPropertySetters_Flow,
)
from fy_library.mixins.property.folder_to_parse.using_setter import (
    FolderToParse_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.fy_py_files_to_parse.using_files_discovery_fy import (
    FyPyFilesToParse_UsingFilesDiscovery_PropertyMixin,
)
from fy_library.mixins.property.mixin_import_map.using_parsed_fy_py_files_fy import (
    MixinImportMap_UsingParsedFyPyFiles_PropertyMixin,
)
from fy_library.mixins.property.parsed_fy_py_files.using_fy_py_parser_fy import (
    ParsedFyPyFiles_UsingFyPyParser_PropertyMixin,
)
from fy_library.mixins.property.parsed_fy_py_files_map_by_key.using_parsed_fy_py_files_fy import (
    ParsedFyPyFilesMapByKey_UsingParsedFyPyFiles_PropertyMixin,
)
from fy_library.mixins.property.project_root_folder.using_setter import (
    ProjectRootFolder_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.required_property_setters_fy_py.using_parsed_fy_py_files_fy import (
    RequiredPropertySettersFyPy_UsingParsedFyPyFiles_PropertyMixin,
)


# fy:start ===>>>
class FyPy_Main_Flow(
    # Property Mixins
    FolderToParse_UsingSetter_PropertyMixin,
    ProjectRootFolder_UsingSetter_PropertyMixin,
    FyPyFilesToParse_UsingFilesDiscovery_PropertyMixin,
    ParsedFyPyFiles_UsingFyPyParser_PropertyMixin,
    ParsedFyPyFilesMapByKey_UsingParsedFyPyFiles_PropertyMixin,
    RequiredPropertySettersFyPy_UsingParsedFyPyFiles_PropertyMixin,
    MixinImportMap_UsingParsedFyPyFiles_PropertyMixin,
    # Base
    FlowBase[None],
):
    def __init__(
        self,
        *args: Any,
        folder_to_parse: Path,
        project_root_folder: Path,
        **kwargs: Any,
    ):
        self._folder_to_parse = folder_to_parse
        self._project_root_folder = project_root_folder
        super().__init__(*args, **kwargs)

    def __call__(self) -> None:
        # fy:end <<<===
        GenerateAndSaveFyPyFiles_UsingParsedFyPyFiles_Flow(
            parsed_fy_py_files=self._parsed_fy_py_files,
            mixin_import_map=self._mixin_import_map,
            parsed_fy_py_files_map_by_key=self._parsed_fy_py_files_map_by_key,
        )()
        GenerateAndSaveFyPyFiles_UsingRequiredPropertySetters_Flow(
            required_property_setters_fy_py=self._required_property_setters_fy_py
        )()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
property fy_py_file_content: str using parsed_fy_py_file:
    property parsed_fy_py_file
    property mixin_imports_code
    property generated_fy_py_code
"""

import abc
from functools import cached_property
from typing import Final

from fy_library.constants import (
    FY_CODE_FILE_END_SIGNATURE,
    FY_PY_FILE_SIGNATURE,
    FY_START_MARKER,
    FY_END_MARKER,
)
from fy_library.mixins.property.generated_fy_py_code.abc_fy import (
    GeneratedFyPyCode_PropertyMixin_ABC,
)
from fy_library.mixins.property.mixin_imports_code.abc_fy import (
    MixinImportsCode_PropertyMixin_ABC,
)
from fy_library.mixins.property.parsed_fy_py_file.abc_fy import (
    ParsedFyPyFile_PropertyMixin_ABC,
)

_NEW_LINE: Final = "\n"


# fy:start ===>>>
class FyPyFileContent_UsingParsedFyPyFile_PropertyMixin(
    # Property_mixins
    ParsedFyPyFile_PropertyMixin_ABC,
    MixinImportsCode_PropertyMixin_ABC,
    GeneratedFyPyCode_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _fy_py_file_content(self) -> str:
        # fy:end <<<===
        stripped_pre_marker_file_content = (
            self._parsed_fy_py_file.pre_marker_file_content.strip()
        )
        fy_py_file_content = (
            f"{self._parsed_fy_py_file.pre_fy_code}"
            f"{FY_PY_FILE_SIGNATURE}"
            f"{self._parsed_fy_py_file.fy_code}"
            f"{FY_CODE_FILE_END_SIGNATURE}\n"
            f"{_NEW_LINE if stripped_pre_marker_file_content else ''}"
            f"{stripped_pre_marker_file_content}"
            f"{_NEW_LINE if stripped_pre_marker_file_content else ''}"
            f"{_NEW_LINE if self._mixin_imports_code else ''}"
            f"{self._mixin_imports_code}"
            f"\n\n{FY_START_MARKER}\n"
            f"{self._generated_fy_py_code}"
            f"{FY_END_MARKER}\n"
            f"{self._parsed_fy_py_file.post_marker_file_content}"
        )

        return fy_py_file_content



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
property method_file_split: MethodFileSplitModel
"""

import abc

from pydantic import BaseModel


class MethodFileSplitModel(BaseModel):
    user_imports: str
    method_name: str
    implementation_name: str
    arguments: str | None
    return_type: str
    mixins: str


# fy:start ===>>>
class MethodFileSplit_PropertyMixin_ABC(abc.ABC):
    @property
    @abc.abstractmethod
    def _method_file_split(self) -> MethodFileSplitModel:
        raise NotImplementedError()
        # fy:end <<<===



# fy:start ===>>>
from typing import Dict


class MixinImportMap_UsingSetter_PropertyMixin:
    @property
    def _mixin_import_map(self) -> Dict[str, str]:
        return self.__mixin_import_map

    @_mixin_import_map.setter
    def _mixin_import_map(self, mixin_import_map: Dict[str, str]) -> None:
        self.__mixin_import_map = mixin_import_map
        # fy:end <<<===



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from typing import List


property import_abstract_method_mixins: List[str] using abstract_method_mixin_and_mixin_import_map:
    property mixin_import_map
    property abstract_method_mixins
"""

from functools import cached_property
from fy_library.mixins.property.abstract_method_mixins.abc_fy import (
    AbstractMethodMixins_PropertyMixin_ABC,
)
from fy_library.mixins.property.mixin_import_map.abc_fy import (
    MixinImportMap_PropertyMixin_ABC,
)
from typing import List
import abc


# fy:start ===>>>
class ImportAbstractMethodMixins_UsingAbstractMethodMixinAndMixinImportMap_PropertyMixin(
    # Property_mixins
    MixinImportMap_PropertyMixin_ABC,
    AbstractMethodMixins_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _import_abstract_method_mixins(self) -> List[str]:
        # fy:end <<<===
        return [
            # method mixins
            self._mixin_import_map[abstract_method_mixin.method_name.snake_case]
            for abstract_method_mixin in self._abstract_method_mixins
        ]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from typing import List


property import_abc: List[str] using constant:
"""

from functools import cached_property
from typing import List


# fy:start ===>>>
class ImportAbc_UsingConstant_PropertyMixin:
    @cached_property
    def _import_abc(self) -> List[str]:
        # fy:end <<<===
        return ["import abc"]



# fy:start ===>>>
from pathlib import Path


class ProjectRootFolder_UsingSetter_PropertyMixin:
    @property
    def _project_root_folder(self) -> Path:
        return self.__project_root_folder

    @_project_root_folder.setter
    def _project_root_folder(self, project_root_folder: Path) -> None:
        self.__project_root_folder = project_root_folder
        # fy:end <<<===



# fy:start ===>>>
class PreMarkerFileContent_UsingSetter_PropertyMixin:
    @property
    def _pre_marker_file_content(self) -> str:
        return self.__pre_marker_file_content

    @_pre_marker_file_content.setter
    def _pre_marker_file_content(self, pre_marker_file_content: str) -> None:
        self.__pre_marker_file_content = pre_marker_file_content
        # fy:end <<<===



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
property jinja2_template_file_name: str using parsed_fy_py_file:
    property parsed_fy_py_file
"""

import abc
from functools import cached_property

from fy_library.domain.parsed_fy_py_file import ParsedFyPyFileKind
from fy_library.mixins.property.parsed_fy_py_file.abc_fy import (
    ParsedFyPyFile_PropertyMixin_ABC,
)


# fy:start ===>>>
class Jinja2TemplateFileName_UsingParsedFyPyFile_PropertyMixin(
    # Property_mixins
    ParsedFyPyFile_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _jinja2_template_file_name(self) -> str:
        # fy:end <<<===
        match self._parsed_fy_py_file.file_type:
            case ParsedFyPyFileKind.FLOW:
                return "flow.jinja2"
            case ParsedFyPyFileKind.METHOD:
                return "method.jinja2"
            case ParsedFyPyFileKind.ABSTRACT_METHOD:
                return "abstract_method.jinja2"
            case ParsedFyPyFileKind.ABSTRACT_PROPERTY:
                return "abstract_property.jinja2"
            case ParsedFyPyFileKind.PROPERTY:
                return "property.jinja2"
            case _:
                raise NotImplementedError(
                    f"No Execution Flow kind for {self._parsed_fy_py_file.file_type}"
                )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from fy_library.domain.fy_py_template_models import BaseTemplateModel


property template_model: BaseTemplateModel using parsed_fy_py_file:
    property parsed_fy_py_file
    property parsed_fy_py_files_map_by_key
"""

import abc
from functools import cached_property

from fy_library.domain.fy_py_template_models import (
    BaseTemplateModel,
    FlowTemplateModelWithPropertySetters,
    FlowTemplateModel,
)
from fy_library.domain.parsed_fy_py_file import ParsedFyPyFileKind
from fy_library.mixins.property.parsed_fy_py_file.abc_fy import (
    ParsedFyPyFile_PropertyMixin_ABC,
)
from fy_library.mixins.property.parsed_fy_py_files_map_by_key.abc_fy import (
    ParsedFyPyFilesMapByKey_PropertyMixin_ABC,
)


# fy:start ===>>>
class TemplateModel_UsingParsedFyPyFile_PropertyMixin(
    # Property_mixins
    ParsedFyPyFile_PropertyMixin_ABC,
    ParsedFyPyFilesMapByKey_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _template_model(self) -> BaseTemplateModel:
        # fy:end <<<===
        if self._parsed_fy_py_file.file_type != ParsedFyPyFileKind.FLOW:
            return self._parsed_fy_py_file.template_model

        assert isinstance(self._parsed_fy_py_file.template_model, FlowTemplateModel)
        property_setters = [
            self._parsed_fy_py_files_map_by_key[
                property_setter.property_name.snake_case
            ].template_model
            for property_setter in self._parsed_fy_py_file.template_model.properties
            if property_setter.implementation_name.snake_case == "setter"
        ]

        return FlowTemplateModelWithPropertySetters.model_validate(
            {
                **self._parsed_fy_py_file.template_model.model_dump(),
                "property_setters": property_setters,
            }
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from typing import List


property filtered_mixin_imports: List[str] using remove_existing_imports:
    property mixin_imports
    property parsed_fy_py_file
"""

import abc
import re
from functools import cached_property
from typing import List, Set, Final

from fy_library.mixins.property.mixin_imports.abc_fy import (
    MixinImports_PropertyMixin_ABC,
)
from fy_library.mixins.property.parsed_fy_py_file.abc_fy import (
    ParsedFyPyFile_PropertyMixin_ABC,
)

_IMPORT_REGEX: Final = re.compile(
    r"^(?P<from>from [\w.]+) import .*$|^(?P<import>import [\w.]+)$", flags=re.DOTALL
)


# fy:start ===>>>
class FilteredMixinImports_UsingRemoveExistingImports_PropertyMixin(
    # Property_mixins
    MixinImports_PropertyMixin_ABC,
    ParsedFyPyFile_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _filtered_mixin_imports(self) -> List[str]:
        # fy:end <<<===
        pre_marker_imports: Set[str] = set()
        for pre_marker_line in self._parsed_fy_py_file.pre_marker_file_content.split(
            "\n"
        ):
            import_regex_result = _IMPORT_REGEX.search(pre_marker_line)
            if import_regex_result is not None:
                pre_marker_imports.add(
                    import_regex_result.group("from")
                    or import_regex_result.group("import")
                )

        mixin_imports_result = []
        for mixin_import in self._mixin_imports:
            import_regex_result = _IMPORT_REGEX.search(mixin_import)
            import_part = import_regex_result.group(
                "from"
            ) or import_regex_result.group("import")
            if import_part not in pre_marker_imports:
                mixin_imports_result.append(mixin_import)

        user_imports_results = []
        for user_import in self._parsed_fy_py_file.user_imports.split("\n"):
            if user_import == "":
                continue
            import_regex_result = _IMPORT_REGEX.search(user_import)
            import_part = import_regex_result.group(
                "from"
            ) or import_regex_result.group("import")
            if import_part not in pre_marker_imports:
                user_imports_results.append(user_import)

        return mixin_imports_result + user_imports_results



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from pathlib import Path
from typing import List


property fy_py_files_to_parse: List[Path] using files_discovery:
    property folder_to_parse
"""

import abc
from functools import cached_property
from pathlib import Path
from typing import List

from fy_library.constants import FY_PY_FILE_SIGNATURE, FY_PY_FILE_EXTENSION
from fy_library.mixins.property.folder_to_parse.abc_fy import (
    FolderToParse_PropertyMixin_ABC,
)


# fy:start ===>>>
class FyPyFilesToParse_UsingFilesDiscovery_PropertyMixin(
    # Property_mixins
    FolderToParse_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _fy_py_files_to_parse(self) -> List[Path]:
        # fy:end <<<===
        fy_py_files_in_directory = list(
            self._folder_to_parse.rglob(f"*{FY_PY_FILE_EXTENSION}")
        )
        fy_py_files: List[Path] = []

        for fy_py_file in fy_py_files_in_directory:
            with open(file=fy_py_file, mode="r") as file:
                while file_line := file.readline():
                    if file_line[0] != "#":
                        break
                if file_line == FY_PY_FILE_SIGNATURE:
                    fy_py_files.append(fy_py_file)
                else:
                    raise SyntaxError(
                        f"File {fy_py_file} does not start with {FY_PY_FILE_SIGNATURE}"
                    )
        return fy_py_files



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from typing import List


property method_mixins_import: List[str] using method_mixins_and_mixin_import_map:
    property method_mixins
    property mixin_import_map
"""

import abc
from functools import cached_property
from typing import List

from fy_library.domain.fy_py_template_models import entity_key
from fy_library.mixins.property.method_mixins.abc_fy import (
    MethodMixins_PropertyMixin_ABC,
)
from fy_library.mixins.property.mixin_import_map.abc_fy import (
    MixinImportMap_PropertyMixin_ABC,
)


# fy:start ===>>>
class MethodMixinsImport_UsingMethodMixinsAndMixinImportMap_PropertyMixin(
    # Property_mixins
    MethodMixins_PropertyMixin_ABC,
    MixinImportMap_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _method_mixins_import(self) -> List[str]:
        # fy:end <<<===
        return [
            # method mixins
            self._mixin_import_map[
                entity_key(
                    mixin_name__snake_case=method_mixin.method_name.snake_case,
                    mixin_implementation_name__snake_case=method_mixin.implementation_name.snake_case,
                )
            ]
            for method_mixin in self._method_mixins
        ]
